/*
 * Copyright (c) 2009 by Gabriel M. Beddingfield <gabriel@teuton.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY, without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <jack/jack.h>
#include <jack/midiport.h>
#include <jack/transport.h>
#include <iostream>
#include <cassert>
#include <cmath>
#include <map>

#include <stdlib.h>
#include <stdint.h>  // e.g. for uint8_t
#include <signal.h>
#include <errno.h>
#include <unistd.h> // for usleep()

using namespace std;

#define APPNAME "midixlate"
#define APPVERSION "0.1.0"

class MidiTranslate
{
public:
    typedef std::map<unsigned char, unsigned char> note_map_t;

    MidiTranslate();
    ~MidiTranslate();

    static int process_callback(jack_nframes_t nframes, void *arg);
    bool good() { return _good; }
    void shutdown() { _good = false; }

private:
    void init_note_map();
    int process(jack_nframes_t nframes);

    bool _good;
    jack_client_t* _client;
    jack_port_t* _inport;
    jack_port_t* _outport;
    note_map_t _note_map;
};

#define CLIENT_FAILURE(msg) {						\
	cerr << "Could not connect to JACK server (" << msg << ")"	\
	     << endl;							\
	if (client) {							\
	    cerr << "...but JACK returned a non-null pointer?" << endl;	\
	    (client) = 0;						\
	}								\
	if (tries) cerr << "...trying again." << endl;			\
    }


#define CLIENT_SUCCESS(msg) {			\
	assert(client);				\
	cout << msg << endl;			\
	tries = 0;				\
    }

MidiTranslate::MidiTranslate() :
    _client(0),
    _inport(0),
    _outport(0),
    _good(true),
    _note_map()
{
    jack_client_t* client;
    jack_port_t* port;
    int rv;
    const char* cname;

    init_note_map();

    jack_status_t status;
    int tries = 2;  // Sometimes jackd doesn't stop and start fast enough.
    while ( tries > 0 ) {
	--tries;
	client = jack_client_open(
	    APPNAME,
	    JackNullOption,
	    &status);
	switch(status) {
	case JackFailure:
	    CLIENT_FAILURE("unknown error");
	    break;
	case JackInvalidOption:
	    CLIENT_FAILURE("invalid option");
	    break;
	case JackNameNotUnique:
	    if (client) {
		cname = jack_get_client_name(client);
		CLIENT_SUCCESS(cname);
	    } else {
		CLIENT_FAILURE("name not unique");
	    }
	    break;
	case JackServerStarted:
	    CLIENT_SUCCESS("JACK Server started for " APPNAME);
	    break;
	case JackServerFailed:
	    CLIENT_FAILURE("unable to connect");
	    break;
	case JackServerError:
	    CLIENT_FAILURE("communication error");
	    break;
	case JackNoSuchClient:
	    CLIENT_FAILURE("unknown client type");
	    break;
	case JackLoadFailure:
	    CLIENT_FAILURE("can't load internal client");
	    break;
	case JackInitFailure:
	    CLIENT_FAILURE("can't initialize client");
	    break;
	case JackShmFailure:
	    CLIENT_FAILURE("unable to access shared memory");
	    break;
	case JackVersionError:
	    CLIENT_FAILURE("client/server protocol version mismatch");
	default:
	    if (status) {
		cerr << "Unknown status with JACK server." << endl;
		if (client) {
		    CLIENT_SUCCESS("Client pointer is *not* null..."
				   " assuming we're OK");
		}
	    } else {
		CLIENT_SUCCESS("Connected to JACK server");
	    }				
	}
    }

    if(client == 0) {
	_good = false;
	return;
    }
    _client = client;

    port = jack_port_register(_client,
			      "midi_out",
			      JACK_DEFAULT_MIDI_TYPE,
			      JackPortIsOutput,
			      0);

    if(port == 0) {
	cerr << "Could not register port." << endl;
	_good = false;
	return;
    }
    _outport = port;

    port = jack_port_register(_client,
			      "midi_in",
			      JACK_DEFAULT_MIDI_TYPE,
			      JackPortIsInput,
			      0);
    if(port == 0) {
	cerr << "Could not register port." << endl;
	_good = false;
	return;
    }
    _inport = port;

    rv = jack_set_process_callback(_client,
				   MidiTranslate::process_callback,
				   static_cast<void*>(this));
    if(rv) {
	cerr << "Could not set process callback." << endl;
	_good = false;
	return;
    }

    rv = jack_activate(_client);
    if(rv) {
	cerr << "Could not activate the client." << endl;
	_good = false;
	return;
    }
	
    return;
}

MidiTranslate::~MidiTranslate()
{
    if(_client) {
	jack_deactivate(_client);
    }
    if(_inport && _client) {
	jack_port_unregister(_client, _inport);
    }
    if(_outport && _client) {
	jack_port_unregister(_client, _outport);
    }
    if(_client) {
	jack_client_close(_client);
    }
    return;
}

void MidiTranslate::init_note_map()
{
    /* Here, add your midi mappings.  If you don't
     * map a note here, then the note will not be translated
     * but pass through unchanged.
     *
     * Syntax:
     *
     * _note_map[orig] = xlate;
     *
     * Where 'orig' is the value of the "Original" note, and
     * 'xlate' is the value of the translated note.
     */
    _note_map[36] = 40;
    _note_map[37] = 36;
    _note_map[38] = 41;
    _note_map[39] = 37;
}

const unsigned char CMD_MASK = 0xF0;
const unsigned char NOTE_ON = 0x90;
const unsigned char NOTE_OFF = 0x80;

int MidiTranslate::process(jack_nframes_t nframes)
{
    void* in_buf = jack_port_get_buffer(_inport, nframes);
    void* out_buf = jack_port_get_buffer(_outport, nframes);
    jack_nframes_t events, k;
    jack_midi_event_t ev;
    unsigned char cmd;
    note_map_t::const_iterator it;
    jack_midi_data_t *odata;
    int rv;

    jack_midi_clear_buffer(out_buf);
    events = jack_midi_get_event_count(in_buf);
    if(events == 0) return 0;

    for(k=0 ; k<events && _good; ++k) {
	rv = jack_midi_event_get(&ev, in_buf, k);
	if( rv != 0 ) {
	    cerr << "Error: Jack said MIDI events were avail., "
		"but didn't deliver." << endl;
	    continue;
	}
	cmd = ev.buffer[0] & CMD_MASK;
	if( cmd == NOTE_ON || cmd == NOTE_OFF ) {
	    it = _note_map.find(ev.buffer[1]);
	    if(it != _note_map.end()) {
		ev.buffer[1] = it->second;
	    }
	}
	odata = jack_midi_event_reserve(out_buf, ev.time, ev.size);
	if(odata) {
	    memcpy((void*)odata,
		   (void*)ev.buffer,
		   ev.size*sizeof(jack_midi_data_t));
	} else {
	    cerr << "MIDI event dropped." << endl;
	}
    }    

    return 0;
}

int MidiTranslate::process_callback(jack_nframes_t nframes, void *arg) {
    return static_cast<MidiTranslate*>(arg)->process(nframes);
}

void about()
{
    cout << APPNAME << " version " << APPVERSION
	 << " Copyright (C) 2009 Gabriel M. Beddingfield" << endl
	 << APPNAME << " comes with ABSOLUTELY NO WARRANTY." << endl
	 << "This is free software, and you are welcome to redistribute it" << endl
	 << "under conditions of the GNU PUBLIC LICENSE (version 2 or later)." << endl
	 << endl;
}

MidiTranslate* jmc_instance = 0;

void sig_handler(int)
{
    if( jmc_instance ) {
	jmc_instance->shutdown();
    }
}

void setup_signal_handler(int signum)
{
    struct sigaction act;
    act.sa_handler = sig_handler;
    sigemptyset(&act.sa_mask);
    act.sa_flags = 0;
    act.sa_restorer = 0;

    if( sigaction( signum, &act, 0 ) ) {
	cerr << "Error:  could not set up signal handler for signal #" << signum << endl;
	switch( errno ) {
	case EFAULT:
	    cerr << "Invalid 'sigaction' structs given to sigaction()" << endl;
	    break;
	case EINVAL:
	    cerr << "Tried to set signal handler for invalid or unauthorized signal" << endl;
	    break;
	default:
	    cerr << "Unknown error #" << errno << endl;
	}	    
	exit(0);
    }
}

int main(int argc, char **argv)
{
    about();

    setup_signal_handler(SIGHUP);
    setup_signal_handler(SIGINT);
    setup_signal_handler(SIGQUIT);
    setup_signal_handler(SIGTERM);

    {
	MidiTranslate jmc;
	jmc_instance = &jmc;

	while(jmc.good())
	{
	    sleep(1);
	}
    }

    cout << "Disconnected from JACK server" << endl;
    return 0;
}
